#
#-#####################################################################
#
# This is Version 2 of the BOX RE-NUMBER / CODE REPLICATION Application. We can now read from and write to the
# clipboard as well as starting the application from the SciTE Tools menu even while another instance of this application is 
# currently running, so you can have two [usually] instances running, 1 from SciTE which closes when the 'Copy To Clipboard'
# button is selected [or the Exit menu button is selected, or the 'X' window button is clicked on] and another started previously
# from the Calc Tools folder.
#
#-#####################################
#
# Starting From the SciTE Tools Menu:
#
#-#####################################
#
# To use this new feature you will need to make some adjustments to the SciTEGlobal.properties file [SciTE Options Tab] in
# the section where we have our custom tools - near the bottom. Here is an example using command number 7 [make sure to
# remove the leading '# ' when copying this]:
#
# command.name.7.*=Box Num/Code Replicator
# command.7.*=C:\ruby\bin\rubyw.exe "\\PDC\FormsProgramming\Shared Code\Calc Tools\tk_box_num_ver2.rbw" "$(FilePath)" "$(SelectionStartColumn)" "$(SelectionEndColumn)" "$(SelectionStartLine)" "$(SelectionEndLine)"
# command.save.before.7.*=1
# command.subsystem.7.*=0
# command.quiet.7.*=0
#
# Steps to follow:
#
# 1) From SciTE, you select [highlight] the code you wish to renumber/replicate.
#
# 2) Either mouse click on the SciTE Tools menu to start this application or use the shortcut keys [i.e. 'Ctrl + 7' using above
#     example].
#	a) Application will start and will be fed the text that you have selected to either renumber or replicate and will
#	   insert the selected text into the input region.
#	b) The Box List for the selected text will be populated using boxes found in the selected text.
#	c) The application window title will be redisplayed with additional text " - STARTED FROM SciTE - " along with the
#	    document id number.
#
# 3) Do the desired renumbering/replicating.
#
# 4) Click on the 'Copy To Clipboard' button.
#	a) The clipboard is cleared and the Output Code [the renumbered/replicated code] is copied into it.
#	b) The application will exit, that is, it will stop executing and will NOT be minimized to the system tray.
#
# 5) In SciTE, you then paste ['Ctrl + v'] the clipboard contents over the text you previously selected if renumbering that
# selection or pasting the clipboard contents where you moved the SciTE cursor to if you wanted to replicate the selected code.
#
# Things to note:
#
#	a) This application automatically saves your SciTE file BEFORE the application runs.
#	b) The application must be eventually closed in order to use the SciTE function Key 'F5' - ONLY if you do not click
#	    on the 'Copy To Clipboard' button.
#
#-#####################################
#
# Using as a stand alone application:
#
#-#####################################
#
# This application DOES NOT need to be run from the SciTE Tools menu, it will still function as a stand alone application.
#
# Steps to follow:
#
# 1) From SciTE, you select [highlight] the code you wish to blank/reverse blank.
#
# 2) Copy the selected code to the clipboard via 'Ctrl + c'.
#
# 3) Either start the application or bring it up from the system tray.
#
# 4) Click on the 'Copy From Clipboard' button.
#	a) The selected SciTE text is inserted into the cleared input region of the tool.
#	b) The Box List for the selected text will be cleared and re-populated with boxes found in the selected text.
#	c) The Exclusion Boxes list will be cleared.
#	d) The First Box, Last Box, Page Offset, and Box Offset fields will be cleared.
#	e) The Repetitions field will be set to 1.
#
# 5) Do the desired renumbering/replicating.
#
# 6) Click on the 'Copy To Clipboard' button.
#	a) The clipboard is cleared and the Output Code [the blanked/reversed blanked code] is copied into it.
#
# 7) Either close or minimze the application.
#
# 8) In SciTE, you then paste ['Ctrl + v'] the clipboard contents over the text you previously selected if renumbering that
# selection or pasting the clipboard contents where you moved the SciTE cursor to if you wanted to replicate the selected code.
#
#-#####################################################################
#

require 'tk'

#
#-#####################################################################
#
# For displaying error messages.
#
require 'Win32API'

#
# Button styles:
#
$MB_OK = 0
$MB_OKCANCEL = 1
$MB_ABORTRETRYIGNORE = 2
$MB_YESNOCANCEL = 3
$MB_YESNO = 4
$MB_RETRYCANCEL = 5

#
# Icon styles:
#
$MB_ICONHAND = 16
$MB_ICONQUESTION = 32
$MB_ICONEXCLAMATION = 48
$MB_ICONASTERISK = 64

#
# Default button:
#
$MB_DEFBUTTON1 = 0
$MB_DEFBUTTON2 = 256
$MB_DEFBUTTON3 = 512

#
# Return values:
#
$IDOK = 1
$IDCANCEL = 2
$IDABORT = 3
$IDRETRY = 4
$IDIGNORE = 5
$IDYES = 6
$IDNO = 7

def messageBox(message, title = 'Message', style = 0)
	mb = Win32API.new("user32", "MessageBox", ['L', 'P', 'P', 'L'], 'L')
	return mb.Call(0, message, title, style)
end

#
#-#####################################################################
#
# Window drawing section:
#
root_title = "BOX RE-NUMBER / CODE REPLICATION Ver 2"

root = TkRoot.new() { title root_title }
root.geometry('+40+200')

top_frame = TkFrame.new(root).pack("side"=>"top", "fill"=>"x")

boxes_offsets_frame = TkFrame.new(top_frame, :borderwidth=>3, :relief=>:ridge).pack(:anchor=>"n", :side=>"left", :pady =>10, :padx=>5 )

boxes_frame = TkFrame.new(boxes_offsets_frame, :borderwidth=>3, :relief=>:ridge).pack(:side=>"left", :padx=>5, :pady=>5)

first_box_frame = TkFrame.new(boxes_frame).pack(:side=>"top", :pady=>5)

first_button = TkButton.new(first_box_frame, :width=>10, :text=>"First Box", :background=>"lightgrey", :foreground=>"blue", "font"=>"arial 10").pack(:side=>"left", :padx=>5)

first_box = TkVariable.new

second_box_frame = TkFrame.new(boxes_frame).pack(:side=>"top")

first_entry = TkEntry.new(second_box_frame, :textvariable=>first_box, :width=>12, "font"=>"arial 10").pack(:side=>"left", :padx=>5)

offsets_frame = TkFrame.new(boxes_offsets_frame, :borderwidth=>3, :relief=>:ridge).pack("side"=>"top", "padx" =>5, "pady" =>5)

first_offset_frame = TkFrame.new(offsets_frame).pack(:side=>"top", :pady=>7)

page_offset_label = TkLabel.new(first_offset_frame, :text=>"Page Offset", :relief=>:sunken, :background=>"lightgrey", :width=>10, "font"=>"arial 10").pack(:side=>"left", :padx=>5)

second_offset_frame = TkFrame.new(offsets_frame).pack("side"=>"top")

page_offset = TkVariable.new
page_offset_entry = TkEntry.new(second_offset_frame, :textvariable=>page_offset, :width=>10, "font"=>"arial 10").pack(:side=>"left", :padx=>5, :pady=>5)

last_button = TkButton.new(first_box_frame, :width=>10, :text=>"Last Box", :background=>"lightgrey", :foreground=>"blue", "font"=>"arial 10").pack(:padx=>5, :side=>"left")

last_box = TkVariable.new
last_entry = TkEntry.new(second_box_frame, :textvariable=>last_box, :width=>12, "font"=>"arial 10").pack(:side=>"left", :padx=>5, :pady=>5)

box_offset_label = TkLabel.new(first_offset_frame, :text=>"Box Offset", :relief=>:sunken, :background=>"lightgrey", :width=>10, "font"=>"arial 10").pack(:side=>"left", :padx=>5)

box_offset = TkVariable.new()
box_offset_entry = TkEntry.new(second_offset_frame, :textvariable=>box_offset, :width=>10, "font"=>"arial 10").pack(:side=>"left", :padx=>5)

clear_boxes_button = TkButton.new(boxes_frame, :width=>10, :text=>"Clear Boxes", :background=>"lightgrey", :foreground=>"blue", "font"=>"arial 10").pack(:padx=>25, :pady=>5, :side=>"left")

clear_boxes_toggle_button = TkButton.new(boxes_frame, :width=>3, :text=>"OFF", :background=>"lightgrey", :foreground=>"blue", "font"=>"arial 10").pack(:pady=>5, :side=>"left")

clear_offsets_button = TkButton.new(offsets_frame, :width=>10, :text=>"Clear Offsets", :background=>"lightgrey", :foreground=>"blue", "font"=>"arial 10").pack(:padx=>17, :pady=>7, :side=>"left")

clear_offsets_toggle_button = TkButton.new(offsets_frame, :width=>3, :text=>"OFF", :background=>"lightgrey", :foreground=>"blue", "font"=>"arial 10").pack(:pady=>7, :side=>"left")

exclusion_box_frame = TkFrame.new(top_frame, :borderwidth=>3, :relief=>:ridge).pack("anchor"=>"n", "side"=>"left", "pady"=>10, :padx=>5)
exclusion_button = TkButton.new(exclusion_box_frame, :background=>"lightgrey", :foreground=>"blue", :text=>"Exclusion Boxes", "font"=>"arial 10").pack(:pady=>5)

exclusion_text_frame = TkFrame.new(exclusion_box_frame).pack("side"=>"top", :pady=>3, :padx=>5)
exclusion_box_bar_vert = TkScrollbar.new(exclusion_text_frame).pack("side"=>"right", "fill"=>"y")
exclusion_box_text = TkText.new(exclusion_text_frame, "wrap"=>"none", "width"=>15, "height"=>6, "font"=>"arial 10").pack("side"=>"left")
exclusion_box_text.yscrollbar(exclusion_box_bar_vert)

clear_exclusion_button = TkButton.new(exclusion_box_frame, :background=>"lightgrey", :foreground=>"blue", :text=> "Clear List", "font"=>"arial 10", "font"=>"arial 10").pack(:pady=>5)

box_list_frame = TkFrame.new(top_frame, :borderwidth=>3, :relief=>:ridge).pack(:side=>"left", :pady=>10, :padx=>10)
box_list_label = TkLabel.new(box_list_frame, :relief=>:sunken, :background=>"lightgrey", :text=>"Box List", "font"=>"arial 10").pack(:fill=>:x)

box_select_frame = TkFrame.new(box_list_frame).pack(:side=>"bottom", "pady" => 5)
box_select_button = TkButton.new(box_select_frame, :background=>"lightgrey", :foreground=>"blue", :text=>"Select Single", "font"=>"arial 10").pack(:side=>"right")

adj_frame = TkFrame.new(box_list_frame).pack("anchor"=>"w", "side"=>"right", "padx" =>10, "pady" =>27)

box_list_bar_vertical = TkScrollbar.new(box_list_frame).pack(:side=>"right", :fill=>"y")
box_list = TkListbox.new(box_list_frame, "background"=>"white", "exportselection" => 0, "width"=>17, "height"=>11, :selectmode=>"extended", "font"=>"arial 10").pack("side"=>"left")
box_list.yscrollbar(box_list_bar_vertical)

adj_page_button = TkButton.new(adj_frame, "width"=>5, :background=>"lightgrey", :foreground=>"blue", :text=> "Page", "font"=>"arial 10").pack("pady" => 10)

adj_box_button = TkButton.new(adj_frame, "width"=>5, :background=>"lightgrey", :foreground=>"blue", :text=>"Box", "font"=>"arial 10").pack()

reset_button = TkButton.new(adj_frame, "width"=>5, :background=>"lightgrey", :foreground=>"blue", :text=>"Reset", "font"=>"arial 10").pack("pady" => 10)

#
#-#####################################################################
#
# Input Code Buttons and Repetition Input Area:
#
renumber_frame = TkFrame.new(root).pack("side"=>"top", "fill"=>"x", :pady =>5)

copy_from_clipboard_button_frame = TkFrame.new(renumber_frame).pack("side"=>"left", "padx"=>10, "pady"=>5)
copy_from_clipboard_button = TkButton.new(copy_from_clipboard_button_frame, "width"=>18, :background=>"lightgrey", :text=>"Copy From Clipboard", :foreground=>"blue", "font"=>"arial 10").pack()

renumber_button_frame = TkFrame.new(renumber_frame).pack("side"=>"left", "padx"=>10)
renumber_button = TkButton.new(renumber_button_frame, :text=>"Re-Number Input Code", "width"=>18, :background=>"lightgrey", :foreground=>"blue", "font"=>"arial 10").pack()

clear_input_button_frame = TkFrame.new(renumber_frame).pack("side"=>"left")
clear_input_button = TkButton.new(clear_input_button_frame, "width"=>18, :text=>"Delete Input Code", :background=>"lightgrey", :foreground=>"blue", "font"=>"arial 10").pack()

get_boxes_button_frame = TkFrame.new(renumber_frame).pack("side"=>"left", "padx"=>10)
get_boxes_button = TkButton.new(get_boxes_button_frame, "width"=>18, :background=>"lightgrey", :foreground=>"blue", :text=>"Create Box List", "font"=>"arial 10").pack()

delete_box_list_button_frame = TkFrame.new(renumber_frame).pack("side"=>"left")
delete_box_list_button = TkButton.new(delete_box_list_button_frame, "width"=>18, :background=>"lightgrey", :foreground=>"blue", :text=>"Delete Box List", "font"=>"arial 10").pack()

reps_frame = TkFrame.new(renumber_frame, :borderwidth=>3, :relief=>:ridge).pack(:side=>"left", :padx =>10)

num_reps_label = TkLabel.new(reps_frame, :text=>"Repetitions", :relief=>:sunken, :background=>"lightgrey", "font"=>"arial 10").pack(:side=>"left")

num_reps = TkVariable.new
num_reps_entry = TkEntry.new(reps_frame, :textvariable=>num_reps, :width=>3, "font"=>"arial 10").pack(:side=>"left")
num_reps.value = "1"

#
#-#####################################################################
#
# Input Code Area:
#
input_frame = TkFrame.new(root).pack("side"=>"top", "fill"=>"both", "expand"=>true)

in_text_frame = TkFrame.new(input_frame).pack("side"=>"left", "fill"=>"both", "padx"=>10, "expand"=>true)
in_text_bar_vert = TkScrollbar.new(in_text_frame).pack("side"=>"right", "fill"=>"y")
in_text = TkText.new(in_text_frame, "width"=>113, "height"=>15, "font"=>"arial 10").pack("side"=>"bottom", "fill"=>"both", "expand"=>true)

in_text.yscrollbar(in_text_bar_vert)

in_text.focus

#
#-#####################################################################
#
# Output Code Area:
#
output_button_frame = TkFrame.new(root).pack("side"=>"top", "fill"=>"both", "expand"=>true)

copy_to_clipboard_button_frame = TkFrame.new(output_button_frame).pack("side"=>"left", "pady"=>5, "padx"=>10)
copy_to_clipboard_button = TkButton.new(copy_to_clipboard_button_frame, "width"=>16, :background=>"lightgrey", :text=>"Copy To Clipboard", :foreground=>"blue", "font"=>"arial 10").pack()

output_label_frame = TkFrame.new(output_button_frame).pack("side"=>"left", "pady"=>5)
output_label = TkLabel.new(output_label_frame, :relief=>:sunken, :background=>"lightgrey", :text=>"Output Code - NOTE: ALL code is selected by default when copying to Clipboard if no code has been selected.", "font"=>"arial 12").pack()

output_frame = TkFrame.new(root).pack("side"=>"top", "fill"=>"both", "expand"=>true)
out_text_frame = TkFrame.new(output_frame).pack("side"=>"left", "padx" => 10, "fill"=>"both", "expand"=>true)

out_text_bar_vert = TkScrollbar.new(out_text_frame).pack("side"=>"right", "fill"=>"y")
out_text = TkText.new(out_text_frame,"width"=>113, "height"=>15, "font"=>"arial 10").pack("side"=>"bottom", "fill"=>"both", "expand"=>true)

out_text.yscrollbar(out_text_bar_vert)

#
#-#####################################################################
#

bottom_frame = TkFrame.new(root).pack("side"=>"top", "fill"=>"x", "pady" => 5)

#
# ###############################################################
#
# Globals:
#
cur_dir = "C:/temp/"

box_array = Array.new

#
# Holds the name and path of the file, set upon start up from SciTE.
#
fname = nil

#
# This tells us if this app was started from the SciTE Tools menu. Set to 'true' in the if (ARGV[0])..end expression
# further down.
#
from_SciTE = false

#
# ###############################################################
#
# Methods:
#
def process_errors(error, var, type)

	e_type = (type.eql?("W") ? "Warning" : "Error")

	err_msg = [
	"  #{e_type.upcase} - Application was started from a SciTE buffer that did not contain the Ruby code from a document.\n",
	"  #{e_type.upcase} - Application was started from a SciTE buffer using invalid SciTEGlobal.properties file arguments.\n\nYou should see the following entry in the SciTEGlobal.properties file for the command number assigned to this application - as an example, here it is for command 7:\n\ncommand.7.*=C:\\ruby\\bin\\rubyw.exe \"\\\\PDC\\FormsProgramming\\Shared Code\\Calc Tools\\tk_box_num_ver2.rbw\" \"$(FilePath)\" \"$(SelectionStartColumn)\" \"$(SelectionEndColumn)\" \"$(SelectionStartLine)\" \"$(SelectionEndLine)\"",
	"  #{e_type.upcase} - process_sub_string( ) has produced a nil string. Application will Abort.\n",
	"  #{e_type.upcase} - Application could not locate the C:\\ruby\\scite\\SciTEGlobal.properties file which is needed to read the tab size value.\n\n You can not invoke this application from SciTE, use instead as a stand alone application where you can first copy your selection to the Clipboard.",
	"  #{e_type.upcase} - Tab size value from C:\\ruby\\scite\\SciTEGlobal.properties #{var} -  Application will default to a tab size value of 8.\n\n If this default tab size does not retrieve the selected text correctly use this application as a stand alone tool where you can copy your selection to the Clipboard.",
	"  #{e_type.upcase} - Application could not locate the tab size value from C:\\ruby\\scite\\SciTEGlobal.properties.\n\n You can not invoke this application from SciTE, use instead as a stand alone application where you first copy your selection to the Clipboard."
	]

	#
	# Using bitwise operations, error is set on a bit by bit basis, where each bit represents an error message.
	#
	if (error > 0)

		mask = 1

		#
		# Math.frexp(error) returns a two element array containing the normalized fraction (a Float) and exponent
		# (a Fixnum) of error. In our case, the fraction is 0.5 and the exponent is a power of 2, as our error is 1 x 2
		# raised to 0,1,2,3,... So this returns 0.5 x 2[a power]. So, say error = 4, which is 1 x 2[2], which is the 3rd
		# error in our list. .frexp returns [0.5, 3] - we want the 3 [3rd error], we use 3 - 1 since we are starting
		# from 0 not 1. Just doing this so that I do not have to keep updating the 0.upto( ) each time I add a new error.
		#
		0.upto(Math.frexp(error)[1]-1) do |shift|
			if (((mask << shift) & error) > 0)
				messageBox(err_msg[shift], "#{e_type}!", $MB_ICONEXCLAMATION|$MB_OK)
			end
		end
	end
end

def create_box_list(box_list, in_text)

	page_num_array = Array.new
	box = String.new
	input_str = String.new

	# Clear the Box List.
	box_list.delete(0, "end")

	# Get input code.
	input_str = in_text.get("1.0", "end")

	input_str.each { |code_line|
		code_line.scan(/(box)(\d+)(_)(\d+)/) {
			page_num_array.push([$2.to_i,$4.to_i]) unless page_num_array.include?([$2.to_i,$4.to_i])
		}
	}

	# Now sort the two-dimensional page_num_array.
	page_num_array.sort!

	# Now put the box parts back together and insert into the Box List widget.
	page_num_array.each_index { |index|
		box_list.insert("end", "box#{page_num_array[index][0]}_#{page_num_array[index][1]}")
	}

	# Select (show highlighting) first list element.
	box_list.selection_set(0)
end

#
# This routine returns a sub-string from the current 'line' read from the file we are selecting from. We call this routine
# from process_current_selection( ), when we are processing either the first 'line' or the last 'line', or when the first
# and last lines are the same line.
#
# First Line Processing: We need to find the index position in 'line' that corresponds to the first_col as supplied from
# SciTE. Note that our index values start from 0 since we are indexing into a string [the supplied 'line' argument]. We
# need to have the tab size to do this. From our end of it, a tab is just a single character just like a space char or the
# letter 'A', but we use the tab size when we process a tab char so we can properly set our 'col' variable which in turn
# allows us to properly set our index values. Our 'col' value corresponds to SciTE column values which start from 1. 
#
# Last Line Processing: We need to find the index position in 'line' that corresponds to the last_col as supplied from
# SciTE. Very similar logic to First Line Processing. We return the sub-string starting from the start of the line and runs
# up to and includes the last index position.
#
# The 'mode' argument tells us if we are processing a sub-string that is on both the first and last line, in which case
# we set the first and last index values, and return the corresponding sub-string.
#
# We look at each byte in 'line', a value of 9 is the tab character.
#
def process_sub_string(mode, line, first_col, last_col, tab_size)

	sub_string = nil
	col = 1
	index = first_index = 0
	tab = tab_size + 1

	line.each_byte do |byte|

		(byte == 9 ? col = tab : col += 1)
		tab += tab_size unless (col < tab)
		index += 1

		case (mode)

			when ("FIRST")

				if (col == first_col)
					sub_string = line[index, line.length - index]
					break
				end

			when ("LAST")

				if (col == last_col)
					sub_string = line[0, index]
					break
				end

			when ("BOTH")

				first_index = index if (col == first_col)

				if (col == last_col)
					sub_string = line[first_index, index - first_index]
					break
				end
		end
	end

	unless (sub_string)

		process_errors(4, nil, "E")
		exit(0)
	end
	sub_string
end

#
# This routine is called when we startup from SciTE.
#
# This retrieves currently selected text from the ruby file. We can not just grab the text via "$(CurrentSelection)"
# [would be used in SciTE Global Properties file] because double quotes are somehow filtered out of the selection so you would 
# get #{box1_1} instead of "#{box1_1}", also, I saw mention somewhere [can not remember exactly where] that
# "$(CurrentSelection)" can only hold so many characters, I think 1000.
#
# Note that we need to retrieve the tab size value from SciTEGlobal.properties file [see the '# Indentation' section in that
# file] to properly interpret the tab chars ["\t"] in the selection we are processing so we grab the correct chars. It is assumed
# that users will not be using either the "Local Options" or the "User Options" files for setting the tab size. Testing has
# revealed that there is a defult tab size of 8 when the tabsize line in the SciTEGlobal.properties file has been commented
# out, or set to 0, regardless of using/not using the other items under the '# Indentation' section.
#
# Note that we only have the starting column of the first line and the stopping column of the last line, so we grab
# entire lines that fall between them. This is exactly what "$(CurrentSelection)" does. So users need to select text
# appropriately.
#
# Returns a string containing the selected Ruby code.
#
def process_current_selection(file, strt_col, end_col, strt_line, end_line, tab_size)

	ln_num = 1
	selection = ''
	process = 0
	same_line = (strt_line == end_line)

	File.open(file, "r") do |src_file|

		while (line = src_file.gets)

			if (ln_num == strt_line)

				if (same_line)

					case
						when ((strt_col == 1) and (end_col == 1)): selection = ''
						when (strt_col == 1)
							selection = process_sub_string("LAST", line, strt_col, end_col, tab_size)
						else
							selection = process_sub_string("BOTH", line, strt_col, end_col, tab_size)
					end
					process = 1 # 1 means we are done, break.
				else
					# Only interested in the starting column since the start and stop lines are not the same.
					selection = (strt_col > 1 ? process_sub_string("FIRST", line, strt_col, end_col, tab_size) : line[0, line.length])

					process = 2 # 2 means do not append to selection since we already did here.
				end
			elsif (ln_num == end_line)

				# We are only interested in the last column since we are on the last line.
				selection += (end_col > 1 ? process_sub_string("LAST", line, strt_col, end_col, tab_size) : '')
				process = 1 # 1 means we are done, break.
			end

			case (process)

				when (1): break
				when (2): process = 3
				when (3): selection += line
			end

			ln_num += 1
		end
	end
	selection
end

#
# Restores to a "clean state" - ready for the next renumbering episode.
#
def clear_all(in_text, out_text, box_list, exclusion_box_text, page_offset, box_offset, num_reps, first_box, last_box)

	in_text.delete("1.0", "end")
	out_text.delete("1.0", "end")
	box_list.delete(0, "end")
	exclusion_box_text.delete("1.0", "end")

	page_offset.value = ''
	box_offset.value = ''

	num_reps.value = 1
	first_box.value = ''
	last_box.value = ''

	# Set the focus to the input text window, and put the cursor at the first char on line one.
	in_text.focus
	in_text.mark_set('insert', '1.0')
end

#
# Called from inside the startup from SciTE section once the existance of this file has been verified - see "if (ARGV[0])"
# block further down below.
#
# This routine reads the "SciTEGlobal.properties" file to retrieve the tab size. We look in the '# Indentation'
# area. We default to a tab size of 8 IF: The 'tabsize' line is commented out, or the value is set to zero - testing
# these cases has revealed that SciTE uses a tab size of 8. The user is given a Warning about this and told they
# can use this application as a stand alone version if the selected text is not as expected, i.e. the default size does
# not work.
#
# Returns:
#
# tab_size set to 'nil' if we could not determine the tab size, else we return a positive tab size value.
#
def get_tab_size(file)

	process = false
	tab_size = nil

	File.open(file, "r") do |src_file|

		while (line = src_file.gets)

			case (line.delete("\s").upcase)
			
				when (/^#INDENTATION$/): process = true

				when (/^#~?TABSIZE=/)

					if (process)

						process_errors(16, "is commented out", "W")
						tab_size = 8
						break
					end

				when (/^TABSIZE=(\d+)$/)

					if (process)

						if ($1.to_i > 0)

							tab_size = $1.to_i
						else
							process_errors(16, "is set to 0", "W")
							tab_size = 8
						end
						break
					end
					
				when (/^#FOLDING$/): break
			end
		end
	end
	tab_size
end

#
# ###############################################################
#
# Menu:
#
#		Exit:
#			Exit application.
#
main_menu = TkMenu.new(root)
main_menu.add('command', :label=> "Exit", :command=> proc { exit(0) })

#
# Adding main_menu to the root window.
#
root.menu(main_menu)

#
# ###############################################################
#
# Button bindings:
#
first_button.bind("Button-1") {
	#~ first_box.value = box_list.get("active")
	box_list.get("active").scan(/box\d+_\d+/) { |box|
		first_box.value = box
	}
}

last_button.bind("Button-1") {
	#~ last_box.value = box_list.get("active")
	box_list.get("active").scan(/box\d+_\d+/) { |box|
		last_box.value = box
	}
}

exclusion_button.bind("Button-1") {
	#~ box_text = box_list.get("active")
	#~ exclusion_box_text.insert("end", box_text + "\n") if (!box_text.empty?)
	box_indices_array = Array.new
	exclusion_box_array = Array.new

	exclusion_box_array = exclusion_box_text.get("1.0", "end").scan(/box\d+_\d+/)

	box_indices_array = box_list.curselection

	box_indices_array.each { |x|
		box_list.get(x).scan(/box\d+_\d+/) { |box|
			exclusion_box_text.insert("end", box + "\n") unless exclusion_box_array.include?(box)
		}
	}

	# Now sort the list and redisplay it.
	exclusion_box_array.clear
	exclusion_box_text.get("1.0", "end").scan(/(box)(\d+)(_)(\d+)/) {
		exclusion_box_array.push([$2.to_i,$4.to_i])
	}

	exclusion_box_array.sort!
	exclusion_box_text.delete("1.0", "end")

	exclusion_box_array.each_index { |index|
		exclusion_box_text.insert("end", "box#{exclusion_box_array[index][0]}_#{exclusion_box_array[index][1]}\n")
	}
}

box_select_button.bind("Button-1") {

	# Determine the current selectmode configuration, it will be either "browse", only one item
	# selectable at a time, or "multiple", where more than one item at a time can be selected.
	# Switch configuration between the two: "browse" or "multiple".
	#
	case box_list.cget("selectmode")

		when "extended"

			box_list.configure("selectmode", "multiple")
			box_select_button.configure("text", "Select Multiple")

		#~ when "browse"

			#~ box_list.configure("selectmode", "multiple")
			#~ box_select_button.configure("foreground", "blue")

		when "multiple"

			# User can select (highlight) multiple list elements at once.
			box_list.configure("selectmode", "extended")
			box_select_button.configure("text", "Select Single")

			# Clear all selected elements, then select (highlight) the last element selected,
			# which is the currently 'active' element - the last one clicked on.
			box_list.selection_clear(0, "end")
			box_list.selection_set(box_list.index("active"))

		#~ when "multiple"
			#~ # User can select (highlight) multiple list elements at once.

			#~ box_list.configure("selectmode", "browse")
			#~ box_select_button.configure("foreground", "black")

			#~ # Clear all selected elements, then select (highlight) the last element selected,
			#~ # which is the currently 'active' element - the last one clicked on.
			#~ box_list.selection_clear(0, "end")
			#~ box_list.selection_set(box_list.index("active"))
	end
}

adj_page_button.bind("Button-1") {

	# Each box can have it's own page number adjustment.
	box_indices_array = Array.new
	box = String.new

	box_indices_array = box_list.curselection
	box_indices_array.each { |x|
		box = box_list.get(x).sub(/(box)(\d+)(_)(\d+)(    )?(-?\d+|P)?(  )?(-?\d+|B)?/) {
			$1+$2+$3+$4+"    #{page_offset.value.to_i.to_s}"+($7 == nil ? "  " : $7)+($8 == nil ? "B" : $8)
			}
		box_list.delete(x)
		box_list.insert(x, box)
		box_list.selection_set(x)
		box_list.activate(x)
	}
}

adj_box_button.bind("Button-1") {

	# Each box can have it's own box number adjustment.
	box_indices_array = Array.new
	box = String.new

	box_indices_array = box_list.curselection
	box_indices_array.each { |x|
		box = box_list.get(x).sub(/(box)(\d+)(_)(\d+)(    )?(-?\d+|P)?(  )?(-?\d+|B)?/) {
			$1+$2+$3+$4+($5 == nil ? "    " : $5)+($6 == nil ? "P" : $6)+"  #{box_offset.value.to_i.to_s}"
			}
		box_list.delete(x)
		box_list.insert(x, box)
		box_list.selection_set(x)
		box_list.activate(x)
	}
}

reset_button.bind("Button-1") {

	# Reset the selected box(es) to just the box portion - ex. 'box1_1'.
	box_indices_array = Array.new
	box = String.new

	box_indices_array = box_list.curselection
	box_indices_array.each { |x|
		box = box_list.get(x).sub(/(box\d+_\d+)(    )?(-?\d+|P)?(  )?(-?\d+|B)?/) { $1 }
		box_list.delete(x)
		box_list.insert(x, box)
		box_list.selection_set(x)
		box_list.activate(x)
	}
}

clear_exclusion_button.bind("Button-1") { exclusion_box_text.delete("1.0", "end") }

clear_input_button.bind("Button-1") {

	clear_all(in_text, out_text, box_list, exclusion_box_text, page_offset, box_offset, num_reps, first_box, last_box)
	}

clear_boxes_button.bind("Button-1") {

	first_box.value = ''
	last_box.value = ''
	}

clear_boxes_toggle_button.bind("Button-1") {

	#
	# User can toggle this "ON/OFF" - clears the First/Last box values when the renumber button is selected.
	#
	case clear_boxes_toggle_button.cget("text")

		when "OFF" then clear_boxes_toggle_button.configure("text", "ON")
		when "ON" then clear_boxes_toggle_button.configure("text", "OFF")
	end
	}

clear_offsets_button.bind("Button-1") {

	page_offset.value = ''
	box_offset.value = ''
	}

clear_offsets_toggle_button.bind("Button-1") {

	#
	# User can toggle this "ON/OFF" - clears the Page/Box offset values when the renumber button is selected.
	#
	case clear_offsets_toggle_button.cget("text")

		when "OFF" then clear_offsets_toggle_button.configure("text", "ON")
		when "ON" then clear_offsets_toggle_button.configure("text", "OFF")
	end
	}

copy_from_clipboard_button.bind("Button-1") {

	#
	# 1) First restore to a "clean state" where all lists and vars are cleared.
	# 2) Copy the contents from the clipboard and set the cursor at the start of the text area.
	# 3) Build the Box List from the contents of the in_text widget.
	#
	clear_all(in_text, out_text, box_list, exclusion_box_text, page_offset, box_offset, num_reps, first_box, last_box)

	in_text.insert("1.0", TkClipboard.get(nil))
	in_text.focus
	in_text.mark_set('insert', '1.0')

	create_box_list(box_list, in_text)
}

renumber_button.bind("Button-1") {

	page_num_array = Array.new
	exclusion_box_array = Array.new
	box_page_num_array = Array.new
	box_array = Array.new

	box = String.new
	input_str = String.new
	output_str = String.new

	#
	# The 'First Box' and 'Last Box' fields are used when you do not want to renumber ALL boxes found in the
	# in_text widget, just the range specified by these boxes, which can span pages. When you make the Box List,
	# you can see that the boxes are sorted from lowest to highest page wise and box number wise per page, so
	# you would basically be saying start the renumbering from one box in the list and stop the renumbering when
	# you reach the last box you selected from the list. The Box List page and box offsets still take precedence
	# over the global offsets ONLY when the box falls in the box range else NO renumbering is done.
	#
	# Check if we are to clear the First/Last Box and Page/Box Offsets values after the renumbering/replicating:
	#
	if (clear_boxes_toggle_button.cget("text") == "ON")

		first_box.value = ''
		last_box.value = ''
	end

	if (clear_offsets_toggle_button.cget("text") == "ON")

		page_offset.value = ''
		box_offset.value = ''
	end

	input_str = in_text.get("1.0", "end")

	# These are the global adjust amounts, used if a particular box does not have it's own values.
	box_adjust = box_offset.to_s.to_i
	page_adjust = page_offset.to_s.to_i

	box_array = box_list.get(0, "end")
	
	box_array.each_index { |index|
		if (box_array[index] =~ /(box\d+_\d+)(    )?(-?\d+|P)?(  )?(-?\d+|B)?/)
			box_array[index] = $1
			page_adj = if ($3 != nil)
						($3 == "P" ? page_adjust : $3.to_i)
					else
						page_adjust
					end
			box_adj = if ($5 != nil)
						($5 == "B" ? box_adjust : $5.to_i)
					else
						box_adjust
					end
			box_page_num_array.push([page_adj, box_adj])
		end
	}

	if (first_box.to_s =~ /(box)(\d+)(_)(\d+)/)
		first_page = $2.to_i
		first_num = $4.to_i
	else
		first_page = 0
		first_num = 0
	end

	if (last_box.to_s =~ /(box)(\d+)(_)(\d+)/)
		last_page = $2.to_i
		last_num = $4.to_i
	else
		last_page = 0
		last_num = 0
	end

	# If both first_box and last_box are empty, run using the boxes in the input text. A sorted list will be
	# generated and the values will come from the top and bottom as first, and last, respectively.
	if ((first_page + first_num + last_page + last_num) == 0)

		input_str.each { |code_line|
			code_line.scan(/(box)(\d+)(_)(\d+)/) {
				page_num_array.push([$2.to_i,$4.to_i]) unless page_num_array.include?([$2.to_i,$4.to_i])
			}
		}

		# Now sort the two-dimensional page_num_array.
		page_num_array.sort!
	
		# Now convert the first pair of values to first_page/first_num values and the last pair to last_page/last/num values,
		# in place of user entered. IF the input_str is empty, use the default 0 values, which changes nothing.
		if (page_num_array.size > 0)
			first_page = page_num_array[0][0]
			first_num = page_num_array[0][1]
			last_page = page_num_array[page_num_array.size - 1][0]
			last_num = page_num_array[page_num_array.size - 1][1]
		end
	end
	
	exclusion_box_array = exclusion_box_text.get("1.0", "end").scan(/box\d+_\d+/)

	1.upto(num_reps.value.to_i) { |x|
		input_str.each { |code_line|

			code_line.gsub!(/(box)(\d+)(_)(\d+)/) { |box|

				page = $2.to_i
				num = $4.to_i

				# Adjust the page and box numbers if the box is not in the exclusion list.
				if (!exclusion_box_array.include?(box))

					adjust = false

					# First check that we are in the page range.
					if ((first_page..last_page) === page)
						if (first_page == last_page)
							# All we need to do is make sure the box num is in the correct range.
							if ((first_num..last_num) === num) then adjust = true end
						else
							# Our box range covers more than one page, need to check if our current
							# box is on the first or last page of the range.
							if (page == first_page)
								if (num >= first_num) then adjust = true end
							elsif (page == last_page)
								if (num <= last_num) then adjust = true end
							else adjust = true
							end
						end

						if (adjust)
							if (box_array.length > 0)
#~ out_text.insert("end", box_list_array[box_list_array.index(box)])

								page += box_page_num_array[box_array.index(box)][0] * x
								num += box_page_num_array[box_array.index(box)][1] * x
							else
								page += page_adjust * x
								num += box_adjust * x
							end
						end
					end
				end

				# Prepare the box token for re-insertion back into the string.
				$1 + page.to_s + $3 + num.to_s
			}
			output_str += code_line
		}
	}

	out_text.delete("1.0", "end")
	out_text.insert("end", output_str)

	#
	# If the in_text text has the newline char on it's own line we want to delete the extra newline from the 
	# out_text that Tk always adds whe we insert - i.e. we replicate that newline char when we do the above
	# renumbering, plus Tk adds it's own newline character when we inserted just above.
	#
	# We 'backup' one character from the end of the text {the 'c' portion of the below "end-1c"} and get the index value
	# of that character [the newline character]. The index is in the form "10.5" as an example, which means line 10, position
	# 5. We split that result on the decimal point and we check if it is positive, which means the newline is not on it's own
	# line, but is at the end of a line of text. split returns a two element array, the line number is the first element, we want
	# to check the second element. If the second element is not positive, then we know we will have that newline [which
	# is on it's own line] PLUS the added newline from our insertion into the out_text, so we want to delete that extra
	# newline here.
	#
	out_text.delete("end-1c") unless (in_text.index("end-1c").split(".")[1].to_i > 0)

	# Set the focus to the output text window, and put the cursor at the first char on line one.
	out_text.focus
	out_text.mark_set('insert', '1.0')
}

get_boxes_button.bind("Button-1") { create_box_list(box_list, in_text) }

delete_box_list_button.bind("Button-1") {

	box_list.delete(0, "end")

	# Clearing the Exclusion Boxes list since it is derived from the Box List.
	exclusion_box_text.delete("1.0", "end")

	# Clearing the First Box and Last Box text vars since they are derived from the Box List.
	first_box.value = ''
	last_box.value = ''
}

copy_to_clipboard_button.bind("Button-1") {

# debug - demonstrating how the "end-1c" and other variants work. "end" is just past the very last character of the line,
# it would be where the next insertion would take place. "end-1c" moves into the string from the end one character position.
#~ out_text.insert("end", "end-1c..end: #{in_text.get("end-1c","end")[0]}\n")
#~ out_text.insert("end", "end-2c..end-1c: #{in_text.get("end-2c","end-1c")[0]}\n")
#~ out_text.insert("end", "end-3c..end-2c: #{in_text.get("end-3c","end-2c")[0]}\n")

	#
	# Copies text from the out_text widget to the clipboard. If no text has been selected it defaults to
	# copying ALL text to the clipboard. If started from the SciTE Tools Menu the application will then exit(0).
	#
	# If no out_text text has been selected, we will do the following when copying all text:
	#
	# If the newline character is not on it's own line in the in_text widget [looking at that since that is what we used for setting
	# the out_text widget] we will copy all out_text minus the newline character since we do not want an extra line in our source
	# file when we Ctrl-V from the clipboard. If however, the newline is on it's own line in the in_text widget, then we do want the
	# newline character.
	#
	if (out_text.tag_ranges("sel").empty?)

		output = out_text.get("1.0", in_text.index("end-1c").split(".")[1].to_i > 0 ? "end - 2c" : "end - 1c")
	else
		output = out_text.get("sel.first", "sel.last")
	end

	#
	# TkClipboard.set first clears the clipboard.
	#
	TkClipboard.set(output.empty? ? '' : output, nil)

	exit(0) if (from_SciTE)
}

#
# ###############################################################
#
# This handles the tool startup from the SciTE tools menu. We first start by verifying if we have Ruby code that
# is associated with a document, else we display an error message and abort. We then verify if we have 5 arguments,
# else we display an error message and abort. We then check if code has been selected [in SciTE], if selected we read
# the selected code and insert it into the input area. The associated code Box List is also populated and we display the
# document ID and a 'started from SciTE' message in the window title area.
# 
if (ARGV[0])

	#
	# We need to ensure this application was started from a document's Ruby code file i.e. has a file name
	# in AAAAxxxx.rbw where A is any alpha character and x is any digit.
	#
	if (File.basename(ARGV[0]) =~ /([a-zA-Z]{4})\d{4}\.rbw/)

		#
		# If set up properly in the SciTE Global Options file, we should have 5 arguments - the file path/name,
		# the start column, the end column, the start line, and the stop line, in that order. Note that we do not
		# actually confirm proper order, just the number of arguments.
		#
		if (ARGV.length == 5)

			#
			# We need to extract the tabsize setting from C:\ruby\scite\SciTEGlobal.properties, down in the 
			# '# Indentation' section. Tab size is used in properly selecting the text in the ruby file for a 
			# document. If for some reason the file does not exist, or can not be found given the supplied path,
			# we display an error message and abort. If we find the file and either the tabsize field is commented
			# out or has a value of 0 we will give a warning message and default to a tabsize of 8 as revealed by
			# testing such a scenario.
			#
			if (File.exist?("C:/ruby/scite/SciTEGlobal.properties"))

				if (tab_size = get_tab_size("C:/ruby/scite/SciTEGlobal.properties"))
# debug:
#~ out_text.insert("end", "ARGV[1].to_i = #{ARGV[1].to_i}\nARGV[2].to_i = #{ARGV[2].to_i}\nARGV[3].to_i = #{ARGV[3].to_i}\nARGV[4].to_i = #{ARGV[4].to_i}\n")
# end of debug
					#
					# Now flag that we started this tool from SciTE so we will automatically shut down when the
					# 'Copy To Clipboard' button has been clicked on.
					#
					from_SciTE = true
			
					#
					# fname is the whole thing - path and file name.
					#
					fname = ARGV[0]
		
					root.configure("title", "#{root_title} - STARTED FROM SciTE - Document ID: #{File.basename(fname, ".*").upcase}")
		
					in_text.insert("end", process_current_selection(fname, ARGV[1].to_i, ARGV[2].to_i, ARGV[3].to_i, ARGV[4].to_i, tab_size))
					create_box_list(box_list, in_text)
				else
					process_errors(32, nil, "E")
					exit(0)
				end
			else
				process_errors(8, nil, "E")
				exit(0)
			end
		else
			#
			# Since we do not have 5 argumants, we shut the app down via exit(0).
			#
			process_errors(2, nil, "E")
			exit(0)
		end
	else
		process_errors(1, nil, "E")

		#
		# We shut the app down.
		#
		exit(0)
	end
end

#
# 03/13/15 [HET]: This no longer pertains, but decided to keep comments and commented code in place for possible future use.
#
#
# This calls exit_app( ) when the user clicks on the 'X' button to close the application window so we
# can prevent the application output from writing/overwriting [when selected text] the SciTE file.
#
# This is because this application can run as a stand alone app in which case we do not automatically write to the
# SciTE file OR it can run from the SciTE Tools menu as a command in which case the app will close upon renumbering/replicating
# or could close when the user Exits the app from the menu or exits from the 'X' button.
#
#~ TkRoot.new.protocol('WM_DELETE_WINDOW', proc{exit_app(from_SciTE)})

Tk.mainloop()
